#!/usr/bin/env python3

from isisdl.backend.crypt import get_credentials
from isisdl.backend.request_helper import CourseDownloader
from isisdl.backend.utils import args, database_helper
from isisdl.bin.update import install_latest_version
from isisdl.settings import is_first_time
from isisdl.version import __version__
import isisdl.bin.config as config


def maybe_print_version_and_exit() -> None:
    if not args.version:
        return

    print(__version__)
    exit(0)


def main() -> None:
    maybe_print_version_and_exit()
    install_latest_version()

    if is_first_time:
        print("It seams as if this is your first time executing isisdl. Welcome <3\n")
        config.main()
        print("\n\nIn the next version I will rediscover your files automatically.\nIf you have already run me before, press CTRL+C and run `isisdl-sync`.")
        # import isisdl.bin.sync_database as sync
        # print("Rediscovering your files ...")
        # sync.main()
        print("\n\nI am now starting to download your files ...\n")

    user = get_credentials()

    dl = CourseDownloader(user)

    dl.start()

    print("\nDone! Have a nice day ^.^")


# TODO:
#   Autolog to server
#   H265
#
#   Maybe systemd timer


if __name__ == "__main__":
    database_helper.get_state()
    main()



#!/usr/bin/env python3
from __future__ import annotations

import os
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import List, Tuple, Dict

from pymediainfo import MediaInfo

from isisdl.backend.crypt import get_credentials
from isisdl.backend.request_helper import RequestHelper, PreMediaContainer
from isisdl.backend.utils import path, calculate_local_checksum, database_helper, calculate_online_checksum_file
from isisdl.settings import course_dir_location, enable_multithread, sync_database_num_threads


def delete_missing_files_from_database() -> None:
    checksums = database_helper.get_checksums_per_course()

    for course in os.listdir(path(course_dir_location)):
        for file in Path(path(course_dir_location, course)).rglob("*"):
            if file.is_file():
                correct_course = checksums[course]
                try:
                    correct_course.remove(calculate_local_checksum(file))
                except KeyError:
                    pass

    for row in checksums.values():
        for item in row:
            database_helper.delete_by_checksum(item)

    num = sum(len(row) for row in checksums.values())
    print(f"Deleted {num} corruped entr{'ies' if num != 1 else 'y'} from the database to be redownloaded.")


def prep_container_and_dump(container: PreMediaContainer, file: Path) -> bool:
    container.location, container.name = os.path.split(file)
    container.checksum = calculate_local_checksum(file)

    return container.dump()


said_text = False


def restore_database_state(helper: RequestHelper) -> None:
    files_to_download: Dict[Path, List[PreMediaContainer]] = defaultdict(list)

    added_num_step_1 = 0

    # TODO: Threads

    for course in helper.courses:
        available_videos = course.download_videos(helper.session)
        available_documents = course.download_documents(helper)
        available_documents.extend(helper.download_mod_assign())

        videos, documents = defaultdict(list), defaultdict(list)
        for item in available_videos:
            videos[item.size].append(item)

        for item in available_documents:
            documents[item.size].append(item)

        def corrupted_file_prompt(file: Path) -> None:
            global said_text
            if said_text is False:
                print("I've found the following corrupted files.\nIf you know that I've downloaded it go ahead and delete them.")
                said_text = True

            print(file.as_posix())

        for file in Path(course.path()).rglob("*"):
            if not os.path.isfile(file):
                continue

            if database_helper.get_name_by_checksum(calculate_local_checksum(file)) is not None:
                continue

            possible: List[PreMediaContainer]
            if os.path.splitext(file.name)[1] == ".mp4":
                info = MediaInfo.parse(file)
                if info.tracks[0].duration is None:
                    corrupted_file_prompt(file)
                    continue

                possible = videos[int(info.tracks[0].duration / 1000)]

            else:
                possible = documents[file.stat().st_size]

            if len(possible) == 0:
                corrupted_file_prompt(file)

            elif len(possible) == 1:
                added_num_step_1 += not prep_container_and_dump(possible[0], file)

            else:
                files_to_download[file].extend(possible)

    def check_multiple(file: Path, containers: List[PreMediaContainer]) -> int:
        checksums: Dict[str, Tuple[int, PreMediaContainer]] = {}
        for item in containers:
            checksum, size = item.calculate_online_checksum(helper.session)
            checksums[checksum] = size, item

        valid_files = []
        for checksum, (size, item) in checksums.items():
            if calculate_online_checksum_file(file, size) == checksum:
                valid_files.append(item)

        if len(valid_files) == 0:
            return 0

        elif len(valid_files) > 1:
            # Two files with same checksum.
            # Since there is no heuristic (aside from filename) that is good enough to differentiate these files
            # they are completely ignored.
            return 0

        prep_container_and_dump(valid_files[0], file)

        return 1

    if enable_multithread:
        with ThreadPoolExecutor(sync_database_num_threads) as ex:
            added_num_step_2 = sum(list(ex.map(check_multiple, *zip(*list(files_to_download.items())))))

    else:
        added_num_step_2 = sum([check_multiple(file, containers) for file, containers in files_to_download.items()])

    recovered_num = added_num_step_1 + added_num_step_2
    total_num = len([item for course in helper.courses for item in Path(course.path()).rglob("*") if item.is_file()])
    if recovered_num == 0:
        print(f"No unrecognized files (checked {total_num})")

    else:
        print(f"I have recovered {recovered_num} / {total_num} possible files.")


def main() -> None:
    user = get_credentials()
    request_helper = RequestHelper(user)

    restore_database_state(request_helper)
    delete_missing_files_from_database()


# TODO: Testing what happens when randomly inserting files

if __name__ == "__main__":
    main()



from __future__ import annotations

import sqlite3
from abc import ABC, abstractmethod
from collections import defaultdict
from threading import Lock
from typing import TYPE_CHECKING, Optional, cast, Set, Dict, List, Tuple, Any

from isisdl.settings import database_file_location, set_database_to_memory

if TYPE_CHECKING:
    from isisdl.backend.request_helper import PreMediaContainer, Course


class SQLiteDatabase(ABC):
    lock: Lock = Lock()

    def __init__(self) -> None:
        from isisdl.backend.utils import path
        if set_database_to_memory:
            self.con = sqlite3.connect(":memory:", check_same_thread=False)
        else:
            self.con = sqlite3.connect(path(database_file_location), check_same_thread=False)

        self.cur = self.con.cursor()

        self.create_default_tables()

    @abstractmethod
    def create_default_tables(self) -> None:
        ...

    def get_state(self) -> Dict[str, List[Any]]:
        res: Dict[str, List[Any]] = {}
        with self.lock:
            names = self.cur.execute("""SELECT name FROM sqlite_master where type = 'table' """).fetchall()
            for name in names:
                res[name[0]] = self.cur.execute(f"""SELECT * FROM {name[0]}""").fetchall()

        return res

    def close_connection(self) -> None:
        self.cur.close()
        self.con.close()


class DatabaseHelper(SQLiteDatabase):
    def create_default_tables(self) -> None:
        with self.lock:
            self.cur.execute("""
                CREATE TABLE IF NOT EXISTS fileinfo
                (name text, file_id text primary key unique, url text, time int, course_id int, checksum text, size int)
            """)

            self.cur.execute("""
                CREATE TABLE IF NOT EXISTS courseinfo
                (name text, id int primary key)
            """)

    def _get_attr_by_equal(self, attr: str, eq_val: str, eq_name: str = "file_id", table: str = "fileinfo") -> Any:
        with DatabaseHelper.lock:
            res = self.cur.execute(f"""SELECT {attr} FROM {table} WHERE {eq_name} = ?""", (eq_val,)).fetchone()

        if res is None:
            return None

        return res[0]

    def get_name_by_checksum(self, checksum: str) -> Optional[str]:
        return cast(Optional[str], self._get_attr_by_equal("name", checksum, "checksum"))

    def get_size_from_file_id(self, file_id: str) -> Optional[int]:
        return cast(Optional[int], self._get_attr_by_equal("size", file_id))

    def get_course_name_and_ids(self) -> List[Tuple[str, int]]:
        with DatabaseHelper.lock:
            return self.cur.execute("""SELECT * FROM courseinfo""").fetchall()

    def delete_by_checksum(self, checksum: str) -> None:
        with DatabaseHelper.lock:
            self.cur.execute("""DELETE FROM fileinfo WHERE checksum = ?""", (checksum,))
            self.con.commit()

    def add_pre_container(self, file: PreMediaContainer) -> bool:
        """
        Returns true iff the element already existed
        """
        with DatabaseHelper.lock:
            already_exists = self.cur.execute("SELECT * FROM fileinfo WHERE checksum = ?", (file.checksum, )).fetchone() is not None

            self.cur.execute("""
                INSERT OR REPLACE INTO fileinfo values (?, ?, ?, ?, ?, ?, ?)
            """, (file.name, file.file_id, file.url, int(file.time.timestamp()), file.course_id, file.checksum, file.size))
            self.con.commit()

            return already_exists

    def add_course(self, course: Course) -> None:
        with DatabaseHelper.lock:
            self.cur.execute("""
                INSERT OR IGNORE INTO courseinfo values (?, ?)
            """, (course.name, course.course_id))
            self.con.commit()

    def get_checksums_per_course(self) -> Dict[str, Set[str]]:
        ret = defaultdict(set)
        with DatabaseHelper.lock:
            for course_name, checksum in self.cur.execute("""SELECT courseinfo.name, checksum from fileinfo INNER JOIN courseinfo on fileinfo.course_id = courseinfo.id""").fetchall():
                ret[course_name].add(checksum)

        return ret

    def delete_file_table(self) -> None:
        with self.lock:
            self.cur.execute("""
                DROP table fileinfo
            """)

        self.create_default_tables()
        pass


class ConfigHelper(SQLiteDatabase):
    def create_default_tables(self) -> None:
        with self.lock:
            self.cur.execute("""
                CREATE TABLE IF NOT EXISTS config
                (key text unique, value text);
            """)

    def _set(self, key: str, value: str) -> None:
        with self.lock:
            self.cur.execute("""
                INSERT OR REPLACE INTO config values (?, ?)
            """, (key, value))
            self.con.commit()

    def _get(self, key: str) -> Optional[str]:
        with self.lock:
            res = self.cur.execute("SELECT value FROM config where key = ?", (key,)).fetchone()
            if res is None:
                return None

            return cast(str, res[0])

    def _delete(self, key: str) -> None:
        with self.lock:
            self.cur.execute("DELETE FROM config where key = ?", (key,))
            self.con.commit()

    def set_user(self, username: str) -> None:
        self._set("username", username)

    def set_clear_password(self, password: str) -> None:
        return self._set("clear_password", password)

    def set_encrypted_password(self, password: str) -> None:
        return self._set("encrypted_password", password)

    def get_user(self) -> Optional[str]:
        return self._get("username")

    def get_clear_password(self) -> Optional[str]:
        return self._get("clear_password")

    def get_encrypted_password(self) -> Optional[str]:
        return self._get("encrypted_password")

    #

    @staticmethod
    def default_filename_scheme() -> str:
        return "0"

    def set_filename_scheme(self, num: str) -> None:
        self._set("filename_scheme", num)

    def get_filename_scheme(self) -> Optional[str]:
        return self._get("filename_scheme")

    def get_or_default_filename_scheme(self) -> str:
        return self.get_filename_scheme() or self.default_filename_scheme()

    def set_throttle_rate(self, num: Optional[str]) -> None:
        if num is None:
            self._delete("throttle_rate")
        else:
            self._set("throttle_rate", num)

    def get_throttle_rate(self) -> Optional[int]:
        value = self._get("throttle_rate")
        if value is None:
            return None

        return int(value)

    #

    @staticmethod
    def default_update_policy() -> str:
        return "2"

    def get_update_policy(self) -> Optional[str]:
        return self._get("update_policy")

    def get_or_default_update_policy(self) -> str:
        return self.get_update_policy() or self.default_update_policy()

    def set_update_policy(self, value: str) -> None:
        self._set("update_policy", value)

    #

    @staticmethod
    def default_telemetry() -> bool:
        return True

    def set_telemetry(self, num: str) -> None:
        self._set("telemetry", num)

    def get_telemetry(self) -> bool:
        value = self._get("telemetry")
        if value is None:
            return self.default_telemetry()

        return value != "0"

    #

    def delete_config(self) -> None:
        with self.lock:
            self.cur.execute("""
                DROP table config
            """)

        self.create_default_tables()



import base64
import getpass
import os
from typing import Optional

from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from isisdl.settings import hash_algorithm, hash_length, hash_iterations, env_var_name_username, env_var_name_password
from isisdl.backend.utils import User, config_helper


def generate_key(password: str) -> bytes:
    # You might notice that the salt is empty. This is a deliberate decision.
    # In this scenario the encrypted file and password file are stored in the same directory.
    # Thus, if a hacker gains access they probably will have access to the salt file as well.

    # As the application is distributed on a local system basis the risc of a data-breach is comparatively low:
    # A hacker might only ever gain access to a single encrypted file.
    salt = b""

    kdf = PBKDF2HMAC(algorithm=hash_algorithm(), length=hash_length, salt=salt, iterations=hash_iterations)

    return base64.urlsafe_b64encode(kdf.derive(password.encode()))


def encryptor(password: str, content: str) -> str:
    key = generate_key(password)
    return Fernet(key).encrypt(content.encode()).decode()


def decryptor(password: str, content: str) -> Optional[str]:
    key = generate_key(password)

    try:
        return Fernet(key).decrypt(content.encode()).decode()
    except InvalidToken:
        return None


def get_credentials() -> User:
    """
    Prioritizes:
        Environment variable > Database > Input
    """

    # First check the environment variables for username *and* password
    username, password = os.getenv(env_var_name_username), os.getenv(env_var_name_password)
    if username is not None and password is not None:
        return User(username, password)

    # Now query the database
    username = config_helper.get_user()
    clear_password = config_helper.get_clear_password()
    encrypted_password = config_helper.get_encrypted_password()

    if username is not None and clear_password is not None:
        return User(username, clear_password)

    if username is not None and encrypted_password is not None:
        user_password = getpass.getpass("Please enter the password you encrypted your password with: ")

        password = decryptor(user_password, encrypted_password)
        if password is None:
            print("Your password is incorrect. Please enter your login information manually (or restart me)!")
        else:
            return User(username, password)

    # If nothing is found prompt the user
    print("Please provide authentication for ISIS.")
    username = input("Username: ")
    password = getpass.getpass("Password: ")

    return User(username, password)



#!/usr/bin/env python3
from __future__ import annotations

import argparse
import atexit
import os
import random
import signal
import string
import sys
from dataclasses import dataclass
from functools import wraps
from pathlib import Path
from queue import PriorityQueue
from typing import Union, Callable, Optional, List, Tuple, Dict, Any, Set, TYPE_CHECKING
from urllib.parse import unquote

import colorama

import isisdl
from isisdl.backend.database_helper import DatabaseHelper, ConfigHelper
from isisdl.settings import working_dir_location, is_windows, settings_file_location, course_dir_location, intern_dir_location, checksum_algorithm, checksum_base_skip, checksum_num_bytes, \
    testing_download_video_size, testing_download_documents_size

if TYPE_CHECKING:
    from isisdl.backend.request_helper import PreMediaContainer


def get_args_main() -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="isisdl", formatter_class=argparse.RawTextHelpFormatter, description="""
    This program downloads all courses from your ISIS page.""")

    parser.add_argument("-V", "--version", help="Print the version number and exit", action="store_true")
    parser.add_argument("-v", "--verbose", help="Enable debug output", action="store_true")
    parser.add_argument("-n", "--num-threads", help="The number of threads which download the content from an individual course.", type=int, default=3)
    parser.add_argument("-d", "--download-rate", help="Limits the download rate to {…}MiB/s", type=float, default=None)
    parser.add_argument("-o", "--overwrite", help="Overwrites all existing files i.e. re-downloads them all.", action="store_true")

    parser.add_argument("-w", "--whitelist", help="A whitelist of course ID's. ", nargs="*")
    parser.add_argument("-b", "--blacklist", help="A blacklist of course ID's. Blacklist takes precedence over whitelist.", nargs="*")

    parser.add_argument("-dv", "--disable-videos", help="Disables downloading of videos", action="store_true")
    parser.add_argument("-dd", "--disable-documents", help="Disables downloading of documents", action="store_true")

    the_args, unknown = parser.parse_known_args()

    if the_args.disable_videos:
        the_args.num_threads *= 2

    course_id_mapping: Dict[str, int] = dict(database_helper.get_course_name_and_ids())

    def add_arg_to_list(lst: Optional[List[Union[str]]]) -> List[int]:
        if lst is None:
            return []

        ret = set()
        for item in lst:
            try:
                ret.add(int(item))
            except ValueError:
                for course, num in course_id_mapping.items():
                    if item.lower() in course.lower():
                        ret.add(int(num))

        return list(ret)

    whitelist: List[int] = []
    blacklist: List[int] = []

    whitelist.extend(add_arg_to_list(the_args.whitelist))
    blacklist.extend(add_arg_to_list(the_args.blacklist))

    the_args.whitelist = whitelist or [True]
    the_args.blacklist = blacklist

    return the_args


def get_args(file: str) -> argparse.Namespace:
    return get_args_main()


def startup() -> None:
    def prepare_dir(p: str) -> None:
        os.makedirs(path(p), exist_ok=True)

    prepare_dir(course_dir_location)
    prepare_dir(intern_dir_location)

    import isisdl

    def restore_link() -> None:
        try:
            os.remove(other_settings_file)
        except OSError:
            pass

        if not is_windows:
            # Sym-linking isn't really supported on Windows / not in a uniform way. Thus, I am not doing that.
            os.symlink(actual_settings_file, other_settings_file)

    actual_settings_file = os.path.abspath(isisdl.settings.__file__)
    other_settings_file = path(settings_file_location)

    if os.path.islink(other_settings_file):
        if os.path.realpath(other_settings_file) != actual_settings_file:
            restore_link()
    else:
        # Damaged link → Either doesn't exist / broken
        restore_link()


def clear() -> None:
    if is_windows:
        os.system('cls')
    else:
        os.system('clear')


def path(*args: str) -> str:
    return os.path.join(working_dir_location, *args)


def sanitize_name(name: str, filename_scheme: Optional[str] = None) -> str:
    # Remove unnecessary whitespace
    name = name.strip()
    name = unquote(name)

    filename_scheme = filename_scheme or config_helper.get_or_default_filename_scheme()

    # First replace umlaute
    for a, b in {"ä": "a", "ö": "o", "ü": "u"}.items():
        name = name.replace(a, b)
        name = name.replace(a.upper(), b.upper())

    # Now start to add chars that don't fit.
    char_string = ""

    if filename_scheme >= "0":
        if is_windows:
            for char in "<>:\"/\\|?*":
                name = name.replace(char, "")

            return name
        else:
            return name.replace("/", "")

    if filename_scheme >= "1":
        # Now replace any remaining funny symbols with a `?`
        name = name.encode("ascii", errors="replace").decode()

        char_string += r"""!"#$%&'()*+,/:;<=>?@[\]^`{|}~"""

    name = name.translate(str.maketrans(char_string, "\0" * len(char_string)))

    # This is probably a suboptimal solution, but it works…
    final = list(name)

    whitespaces = set(string.whitespace + "_")
    i = 0
    while i < len(final):
        char = final[i]

        if char == "\0" or char in whitespaces:
            final.pop(i)

            if char in whitespaces:
                if i < len(final):
                    final[i] = final[i].upper()

            continue

        i += 1

    return "".join(final)


def get_input(message: str, allowed: Set[str]) -> str:
    while True:
        choice = input(message)
        if choice in allowed:
            break

        print(f"\nUnhandled character: {choice!r} is not in the expected: {allowed}.\nPlease try again:\n\n")

    return choice


class OnKill:
    _funcs: PriorityQueue[Tuple[int, Callable[[], None]]] = PriorityQueue()
    _min_priority = 0
    _already_killed = False

    def __init__(self) -> None:
        signal.signal(signal.SIGINT, OnKill.exit)
        signal.signal(signal.SIGABRT, OnKill.exit)
        signal.signal(signal.SIGTERM, OnKill.exit)

    @staticmethod
    def add(func: Any, priority: Optional[int] = None) -> None:
        if priority is None:
            # Generate a new priority → max priority
            priority = OnKill._min_priority - 1

        OnKill._min_priority = min(priority, OnKill._min_priority)

        OnKill._funcs.put((priority, func))

    @staticmethod
    @atexit.register
    def exit(sig: Optional[int] = None, frame: Any = None) -> None:
        if OnKill._already_killed and sig is not None:
            print("Alright, stay calm. I am skipping cleanup and exiting!")
            print("I will redownload the files that are partially downloaded.")

            os._exit(sig)

        if sig is not None:
            sig = signal.Signals(sig)
            if isisdl.backend.request_helper.downloading_files:
                OnKill._already_killed = True
            else:
                os._exit(sig.value)

        for _ in range(OnKill._funcs.qsize()):
            OnKill._funcs.get_nowait()[1]()


def on_kill(priority: Optional[int] = None) -> Callable[[Any], Any]:
    def decorator(function: Any) -> Any:
        # Expects the method to have *no* args
        @wraps(function)
        def _impl(*_: Any) -> Any:
            return function()

        OnKill.add(_impl, priority)
        return _impl

    return decorator


# Shared between modules.
@dataclass
class User:
    username: str
    password: str

    @staticmethod
    def sanitize_name(name: str) -> str:
        # uwu
        if name == "".join(chr(item) for item in [109, 97, 116, 116, 105, 115, 51, 52, 48, 51]):
            return "".join(chr(item) for item in [101, 109, 105, 108, 121, 51, 52, 48, 51])

        return name

    @property
    def sanitized_username(self) -> str:
        return self.sanitize_name(self.username)

    def __repr__(self) -> str:
        return f"{self.sanitized_username}: {self.password}"

    def __str__(self) -> str:
        return f"\"{self.sanitized_username}\""


def calculate_local_checksum(filename: Path) -> str:
    sha = checksum_algorithm()
    sha.update(str(os.path.getsize(filename)).encode())
    curr_char = 0
    with open(filename, "rb") as f:
        i = 1
        while True:
            f.seek(curr_char)
            data = f.read(checksum_num_bytes)
            curr_char += checksum_num_bytes
            if not data:
                break
            sha.update(data)

            curr_char += checksum_base_skip ** i
            i += 1

    return sha.hexdigest()


def calculate_online_checksum_file(file: Path, size: int) -> str:
    chunk = b""
    with file.open("rb") as f:
        while len(chunk) < size:
            chunk += f.read(size - len(chunk))

    return checksum_algorithm(chunk + str(size).encode()).hexdigest()


# Copied and adapted from https://stackoverflow.com/a/63839503
class HumanBytes:
    @staticmethod
    def format(num: Union[int, float]) -> Tuple[float, str]:
        """
        Human-readable formatting of bytes, using binary (powers of 1024) representation.

        Note: num > 0
        """

        unit_labels = ["  B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"]
        last_label = unit_labels[-1]
        unit_step = 1024
        unit_step_thresh = unit_step - 0.5

        unit = None
        for unit in unit_labels:
            if num < unit_step_thresh:
                # Only return when under the rounding threshhold
                break
            if unit != last_label:
                num /= unit_step

        return num, unit


def _course_downloader_transformation(pre_containers: List[PreMediaContainer]) -> List[PreMediaContainer]:
    possible_videos: List[PreMediaContainer] = []
    possible_documents: List[PreMediaContainer] = []

    # Get a random sample of lower half
    video_containers = sorted([item for item in pre_containers if item.is_video], key=lambda x: x.size)
    video_containers = video_containers[:int(len(video_containers) / 2)]

    document_containers = [item for item in pre_containers if not item.is_video]

    random.shuffle(video_containers)
    random.shuffle(document_containers)

    def maybe_add(lst: List[Any], file: PreMediaContainer, max_size: int) -> bool:
        maybe_new_size = sum(item.size for item in lst) + file.size
        if maybe_new_size > max_size:
            return True

        lst.append(file)
        return False

    # Select videos such that the total number of seconds does not overflow.
    for item in video_containers:
        if maybe_add(possible_videos, item, testing_download_video_size):
            break

    for item in document_containers:
        if maybe_add(possible_documents, item, testing_download_documents_size):
            break

    ret = possible_videos + possible_documents
    random.shuffle(ret)
    return ret


startup()
OnKill()
database_helper = DatabaseHelper()
config_helper = ConfigHelper()

args = get_args(os.path.basename(sys.argv[0]))

# Windows specific color codes…
colorama.init()



#!/usr/bin/env python3
import os
import re
import subprocess
import sys
from tempfile import TemporaryDirectory
from typing import Optional

import requests

from isisdl.backend.utils import config_helper
from isisdl.version import __version__


def check_pypi_for_version() -> Optional[str]:
    # Inspired from https://pypi.org/project/pypi-search
    to_search = requests.get("https://pypi.org/project/isisdl/").text
    version = re.search("<h1 class=\"package-header__name\">\n *(.*)?\n *</h1>", to_search)

    if version is None:
        return None

    groups = version.groups()
    if groups is None or len(groups) != 1:
        return None

    return groups[0].split()[1]


def check_github_for_version() -> Optional[str]:
    badge = requests.get("https://github.com/Emily3403/isisdl/actions/workflows/tests.yml/badge.svg").text
    if "passing" not in badge:
        return None

    res = requests.get("https://raw.githubusercontent.com/Emily3403/isisdl/main/src/isisdl/version.py")
    if not res.ok:
        return None

    version = re.match("__version__ = \"(.*)?\"", res.text)
    if version is None:
        return None

    return version.group(1)


def install_latest_version() -> None:
    version_github = check_github_for_version()
    version_pypi = check_pypi_for_version()

    update_policy = config_helper.get_or_default_update_policy()
    if update_policy == "0":
        return

    correct_version = version_github if update_policy == "1" else version_pypi

    if correct_version is None:
        return

    if correct_version > __version__:
        print(f"\nThere is a new version available: {correct_version} (current: {__version__}).")
        print("According to your update policy I will auto-install it.\n")

    else:
        return

    if update_policy == "1":
        old_dir = os.getcwd()
        with TemporaryDirectory() as tmp:
            os.chdir(tmp)
            print(f"Cloning the repository into {tmp} ...")
            ret = subprocess.check_call(["git", "clone", "https://github.com/Emily3403/isisdl"])
            if ret:
                print(f"Cloning failed with exit code {ret}")
                return

            print("Installing with pip ...")
            subprocess.check_call([sys.executable, "-m", "pip", "install", "./isisdl"])
            os.chdir(old_dir)

    if update_policy == "2":
        subprocess.call([sys.executable, "-m", "pip", "install", "--upgrade", "isisdl"])



import os
import platform
import sys
from hashlib import sha256
from typing import TYPE_CHECKING

from cryptography.hazmat.primitives.hashes import SHA3_512

if TYPE_CHECKING:
    pass

# In this file you will find various constants that dictate how isisdl works.
# First up there are things that you may want to change.
# In the second part you should only change stuff if you know what you are doing.

# < Directory options >

# The directory where everything lives in.
# Note: If you want to expand your "~" use `os.path.expanduser("~")`. Otherwise a Directory with the literal `~` will be created in the current working directory.
working_dir_location = os.path.join(os.path.expanduser("~"), "isisdl_downloads")

# The directory where the courses are listed.
# In the courses house all the files which are downloadable.
course_dir_location = "Courses"

# The directory of the database and other persistent functionality.
intern_dir_location = ".intern"

# Will create a symlink to the settings file in the intern dir.
settings_file_location = os.path.join(intern_dir_location, "settings.py")

# Will create the corresponding SQlite Database
database_file_location = os.path.join(intern_dir_location, "state.db")

# Creates a mock database to play on. Done with :memory:
set_database_to_memory = False

# </ Directory options >

# < Checksums >

# A checksum is calculated with this algorithm
checksum_algorithm = sha256

# The number of bytes sampled
checksum_num_bytes = 1024 * 4

# Skips $`checksum_base_skip` ^ i$ bytes per calculation → O(log(n)) time :O
checksum_base_skip = 2

# Number of threads to use for the database requests
sync_database_num_threads = 32

# </ Checksums >


# < Password / Cryptography options >

# Beware: Changing any of these options means loosing compatibility with the old password file.

# This is what Django recommends as of January 2021 (https://github.com/django/django/blob/main/django/contrib/auth/hashers.py#L274)
hash_algorithm = SHA3_512
hash_iterations = 390_000
hash_length = 32

# < Password / Cryptography options >

# < Status options >

# The number of places the progress bar has.
progress_bar_resolution = 10

# Chop off the last ↓ characters of the status message for a ...
status_chop_off = 2

# The refresh time for the status message
status_time = 0.25

# </ Status options >


# < Miscellaneous options >

# If you want to use username and password set these variables accordingly.
env_var_name_username = "ISISDL_USERNAME"
env_var_name_password = "ISISDL_PASSWORD"

# Output os.
config_clear_screen = True

# </ Miscellaneous options >

# Begin second part.
# Don't change anything below this otherwise it might have negative implications. Or do… I don't care :D

# < Download options >
enable_multithread = True

# Sets the chunk size for a download.
download_chunk_size = 2 ** 16

# When ISIS is complaining that you are downloading too fast (Connection Aborted) ↓ s are waited.
sleep_time_for_isis = 3

# Will retry downloading an url ↓ times. If it fails, that MediaContainer will not get downloaded.
num_tries_download = 5

# Will fail a download if ISIS is not responding in ↓ amount of s
download_timeout = 6

# Adds `download_timeout_multiplier ** (0.5 * i)` of timeout every iteration
download_timeout_multiplier = 2

# </ Download options >

# < Miscellaneous options >

# A constant to detect if you are on Windows.
is_windows = platform.system() == "Windows"

# Check if the user is executing the library for the first time → state.db should be missing
is_first_time = not os.path.exists(os.path.join(working_dir_location, database_file_location))

# DownloadThrottler refresh rate in s
token_queue_refresh_rate = 0.1

# Collect the amount of handed out tokens in the last ↓ secs
token_queue_download_refresh_rate = 3

# Yes, changing behaviour when tested is evil. But I'm doing so in order to protect my `~/isisdl_downloads` directory.
is_testing = "pytest" in sys.modules
if is_testing:
    _working_dir_location = working_dir_location
    working_dir_location = os.path.join(os.path.expanduser("~"), "test_isisdl")
    status_time = 2
    download_timeout = 3

# This number represent seconds of video.    (ISIS does not offer a better "size" api…)
testing_download_video_size = 3600 * 1

# Number of bytes downloaded for documents.
testing_download_documents_size = 1_000_000_000

# </ Miscellaneous options >



from __future__ import annotations

import os
import re
import threading
import time
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
from datetime import datetime
from threading import Thread
from typing import Optional, Dict, List, Any, cast, Tuple
from urllib.parse import urlparse, urljoin

from isisdl.backend.downloads import SessionWithKey, MediaType, MediaContainer, Status, DownloadThrottler
from isisdl.backend.utils import User, path, sanitize_name, args, on_kill, database_helper, _course_downloader_transformation
from isisdl.settings import course_dir_location, enable_multithread, checksum_algorithm, is_testing, checksum_num_bytes


@dataclass
class PreMediaContainer:
    name: str
    file_id: str
    url: str
    location: str
    time: datetime
    course_id: int
    is_video: bool
    size: int = -1
    checksum: Optional[str] = None

    @classmethod
    def from_api(cls, file_dict: Dict[str, Any], file_id: str, course: Course) -> PreMediaContainer:
        file_id = f"{file_id}_{checksum_algorithm(file_dict['fileurl'].encode()).hexdigest()}"

        return cls.from_course(file_dict["filename"], file_id, file_dict["fileurl"], course, file_dict["timemodified"], file_dict["filepath"], file_dict["filesize"])

    @classmethod
    def from_course(cls, name: str, file_id: str, url: str, course: Course, last_modified: int, relative_location: str = "", size: int = -1) -> PreMediaContainer:
        # Sanitize bad names
        relative_location = relative_location.strip("/")
        if relative_location:
            if database_helper.get_size_from_file_id(file_id) is None:
                os.makedirs(course.path(sanitize_name(relative_location)), exist_ok=True)

        is_video = "mod/videoservice/file.php" in url
        if is_video:
            relative_location = os.path.join(relative_location, "Videos")

        sanitized_url = urljoin(url, urlparse(url).path)

        location = course.path(sanitize_name(relative_location))
        time = datetime.fromtimestamp(last_modified)

        if "webservice/pluginfile.php" not in url and "mod/videoservice/file.php" not in url:
            # TODO: Server
            pass

        return cls(name, file_id, sanitized_url, location, time, course.course_id, is_video, size)

    def dump(self) -> bool:
        assert self.checksum is not None
        return database_helper.add_pre_container(self)

    def calculate_online_checksum(self, s: SessionWithKey) -> Tuple[str, int]:
        while True:
            running_download = s.get_(self.url, params={"token": s.token}, stream=True)

            if running_download is None:
                continue

            if not running_download.ok:
                assert False

            break

        chunk = running_download.raw.read(checksum_num_bytes, decode_content=True)
        size = len(chunk)

        return checksum_algorithm(chunk + str(size).encode()).hexdigest(), size

    def __hash__(self) -> int:
        return self.file_id.__hash__()


@dataclass
class Course:
    name: str
    course_id: int

    @classmethod
    def from_dict(cls, info: Dict[str, Any]) -> Course:
        name = cast(str, info["displayname"])
        id = cast(int, info["id"])

        return cls(name, id)

    def __post_init__(self) -> None:
        self.prepare_dirs()

    def prepare_dirs(self) -> None:
        for item in MediaType.list_dirs():
            os.makedirs(self.path(item), exist_ok=True)

    def download_videos(self, s: SessionWithKey) -> List[PreMediaContainer]:
        if args.disable_videos:
            return []

        url = "https://isis.tu-berlin.de/lib/ajax/service.php"
        # Thank you isia-tub for this data <3
        video_data = [{
            "index": 0,
            "methodname": "mod_videoservice_get_videos",
            "args": {"coursemoduleid": 0, "courseid": self.course_id}
        }]

        videos_res = s.get_(url, params={"sesskey": s.key}, json=video_data)
        if videos_res is None:
            return []

        videos_json = videos_res.json()[0]

        if videos_json["error"]:
            return []

        return [PreMediaContainer.from_course(item["title"].strip() + item["fileext"], item["id"], item["url"], self, item["timecreated"], size=item["duration"])
                for item in videos_json["data"]["videos"]]

    def download_documents(self, helper: RequestHelper) -> List[PreMediaContainer]:
        if args.disable_documents:
            return []

        content = helper.post_REST("core_course_get_contents", {"courseid": self.course_id})
        if content is None:
            return []

        content = cast(List[Dict[str, Any]], content)

        all_content: List[PreMediaContainer] = []
        for week in content:
            file: Dict[str, Any]
            for file in week["modules"]:
                if "url" not in file:
                    continue

                url: str = file["url"]
                # This is a definite blacklist on stuff we don't want to follow.
                ignore = re.match(
                    ".*mod/(?:"
                    "forum|url|choicegroup|assign|videoservice|feedback|choice|quiz|glossary|questionnaire|scorm|etherpadlite|lti|h5pactivity|"
                    "page"
                    ")/.*", url
                )

                if ignore is not None:
                    # Blacklist hit
                    continue

                if re.match(".*mod/(?:folder|resource)/.*", url) is None:
                    # TODO: server
                    pass

                if "contents" not in file:
                    # TODO: Server
                    continue

                prev_len = len(all_content)
                if "contents" in file:
                    for item in file["contents"]:
                        all_content.append(PreMediaContainer.from_api(item, file["id"], self))

                if len(all_content) == prev_len:
                    known_bad_urls = {
                        "https://isis.tu-berlin.de/mod/folder/view.php?id=1145174"
                    }

                    if url not in known_bad_urls:
                        # TODO: Server
                        pass

        return all_content

    def path(self, *args: str) -> str:
        """
        Custom path function that prepends the args with the `download_dir` and course name.
        """
        return path(course_dir_location, sanitize_name(self.name), *args)

    @property
    def ok(self) -> bool:
        if args.whitelist != [True] and self in args.whitelist:
            return True

        return self in args.whitelist and self not in args.blacklist

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{self.name} ({self.course_id})"

    def __eq__(self, other: Any) -> bool:
        if other is True:
            return True

        if other.__class__ == self.__class__:
            return bool(self.course_id == other.course_id)

        if isinstance(other, (str, int)):
            return str(self.course_id) == str(other)

        return False

    def __hash__(self) -> int:
        return self.course_id


@dataclass
class RequestHelper:
    user: User
    session: SessionWithKey = field(init=False)
    courses: List[Course] = field(default_factory=lambda: [])
    course_id_mapping: Dict[int, Course] = field(default_factory=lambda: {})
    _meta_info: Dict[str, str] = field(default_factory=lambda: {})

    def __post_init__(self) -> None:
        self.session = SessionWithKey.from_scratch(self.user)

        self._meta_info = cast(Dict[str, str], self.post_REST("core_webservice_get_site_info"))
        self._get_courses()

        if args.verbose:
            print("I am downloading the following courses:\n" + "\n".join(item.name for item in self.courses))

    def _get_courses(self) -> None:
        res = cast(List[Dict[str, str]], self.post_REST("core_enrol_get_users_courses", {"userid": self.userid}))
        for item in res:
            course = Course.from_dict(item)

            self.course_id_mapping.update({course.course_id: course})
            database_helper.add_course(course)

            if course.ok:
                self.courses.append(course)
                if not os.path.exists(course.path()):
                    os.makedirs(course.path(), exist_ok=True)

    def post_REST(self, function: str, data: Optional[Dict[str, Any]] = None, params: Optional[Dict[str, Any]] = None) -> Optional[Any]:
        data = data or {}

        data.update({
            "moodlewssettingfilter": "true",
            "moodlewssettingfileurl": "true",
            "moodlewsrestformat": "json",
            "wsfunction": function,
            "wstoken": self.session.token
        })

        url = "https://isis.tu-berlin.de/webservice/rest/server.php"

        response = self.session.post_(url, data=data, params=params)

        if response is None or not response.ok:
            return None

        return response.json()

    def download_content(self) -> List[PreMediaContainer]:
        def download_videos(ret: List[PreMediaContainer], course: Course, session: SessionWithKey) -> None:
            ret.extend(course.download_videos(session))

        def download_documents(ret: List[PreMediaContainer], course: Course) -> None:
            ret.extend(course.download_documents(self))

        def download_mod_assign(ret: List[PreMediaContainer]) -> None:
            ret.extend(self.download_mod_assign())

        if enable_multithread:
            collect: List[Tuple[List[PreMediaContainer], List[PreMediaContainer]]] = [([], []) for _ in range(len(self.courses))]
            threads = []

            mod_assign: List[PreMediaContainer] = []
            mod_assign_thread = Thread(target=download_mod_assign, args=(mod_assign,))
            for col, course in zip(collect, self.courses):
                threads.append(Thread(target=download_documents, args=(col[0], course)))
                threads.append(Thread(target=download_videos, args=(col[1], course, self.session)))

            mod_assign_thread.start()
            for thread in threads:
                thread.start()

            for thread in threads:
                thread.join()

            mod_assign_thread.join()

            documents, videos = [], []
            for item in collect:
                documents.extend(item[0])
                videos.extend(item[1])

        else:
            mod_assign = self.download_mod_assign()
            documents, videos = [], []
            for course in self.courses:
                documents.extend(course.download_documents(self))
                videos.extend(course.download_videos(self.session))

        def sort(lst: List[PreMediaContainer]) -> List[PreMediaContainer]:
            return sorted(lst, key=lambda x: x.time, reverse=True)

        # Download the newest files first
        return sort(documents) + sort(mod_assign) + sort(videos)

    def download_mod_assign(self) -> List[PreMediaContainer]:
        if args.disable_documents:
            return []

        all_content = []
        _assignments = self.post_REST('mod_assign_get_assignments')
        if _assignments is None:
            return []

        assignments = cast(Dict[str, Any], _assignments)

        allowed_ids = {item.course_id for item in self.courses}
        for course in assignments["courses"]:
            if course["id"] in allowed_ids:
                for assignment in course["assignments"]:
                    for file in assignment["introattachments"]:
                        file["filepath"] = assignment["name"]
                        all_content.append(PreMediaContainer.from_api(file, assignment["id"], self.course_id_mapping[course["id"]]))

        # TODO: Also get the submission contents.
        return all_content

    @property
    def userid(self) -> str:
        return self._meta_info["userid"]


def check_for_conflicts_in_files(files: List[PreMediaContainer]) -> None:
    conflicts = defaultdict(list)
    for item in files:
        conflicts[item.name].append(item)

    items: List[List[PreMediaContainer]] = [sorted(item, key=lambda x: x.time if x.time is not None else -1) for item in conflicts.values() if len(item) != 1]
    for row in items:
        locations: Dict[str, List[PreMediaContainer]] = defaultdict(list)
        for item in row:
            locations[item.location].append(item)

        locations = {k: v for k, v in locations.items() if len(v) > 1}

        # TODO: Maybe filter out duplicates by file size

        # TODO: Server

        for new_row in locations.values():
            for i, item in enumerate(new_row):
                basename, ext = os.path.splitext(item.name)
                item.name = basename + f"({i}-{len(new_row) - 1})" + ext


class CourseDownloader:
    helper: Optional[RequestHelper] = None

    def __init__(self, user: User):
        self.user = user

    def start(self) -> None:
        self.make_helper()

        pre_containers = self.build_files()

        check_for_conflicts_in_files(pre_containers)

        if is_testing:
            pre_containers = _course_downloader_transformation(pre_containers)

        media_containers = self.make_files(pre_containers)
        global downloading_files
        downloading_files = media_containers

        # Make the runner a thread in case of a user needing to exit the program → downloading is done in the main thread
        global status
        throttler = DownloadThrottler()
        status = Status(len(media_containers), args.num_threads, throttler)
        downloader = Thread(target=self.download_files, args=(media_containers, throttler))

        downloader.start()
        status.start()

        downloader.join()
        status.join(0)

    def make_helper(self) -> None:
        self.helper = RequestHelper(self.user)

    def build_files(self) -> List[PreMediaContainer]:
        assert self.helper is not None

        return self.helper.download_content()

    def make_files(self, files: List[PreMediaContainer]) -> List[MediaContainer]:
        assert self.helper is not None

        new_files = [MediaContainer.from_pre_container(file, self.helper.session) for file in files]
        filtered_files = [item for item in new_files if item is not None]

        return filtered_files

    def download_files(self, files: List[MediaContainer], throttler: DownloadThrottler) -> None:
        def download(file: MediaContainer) -> None:
            assert status is not None
            if enable_multithread:
                thread_id = int(threading.current_thread().name.split("T_")[-1])
            else:
                thread_id = 0

            status.add(thread_id, file)
            file.download(throttler)
            status.finish(thread_id)

        if enable_multithread:
            with ThreadPoolExecutor(args.num_threads, thread_name_prefix="T") as ex:
                list(ex.map(download, files))
        else:
            for file in files:
                download(file)

    @staticmethod
    @on_kill()
    def shutdown_running_downloads(*_: Any) -> None:
        if downloading_files is None:
            return

        for item in downloading_files:
            item.stop()

        if status is not None:
            status.shutdown()

        # Now wait for the downloads to finish
        while not all(item.done for item in downloading_files):
            time.sleep(0.25)


status: Optional[Status] = None
downloading_files: Optional[List[MediaContainer]] = None



#!/usr/bin/env python3
import sys
from getpass import getpass
from typing import List, Tuple, Optional, Union

from isisdl.backend.crypt import encryptor
from isisdl.backend.utils import config_helper, get_input, User, path, clear
from isisdl.settings import is_first_time, is_windows, is_testing, database_file_location, config_clear_screen

explanation_depth = "2"
indent = "    "


def generic_prompt(question: str, values: List[Tuple[str, str, str]], default: int, overwrite_output: Optional[str] = None, allow_stored: Optional[Union[str, int]] = None) -> str:
    if overwrite_output:
        return overwrite_output

    if config_clear_screen:
        clear()

    print(question + "\n")
    for i, (val, tldr, detail) in enumerate(values):
        print(f"{indent}{i}. {val}{' [default]' if i == default else ''}")
        if explanation_depth > "0":
            if tldr:
                print()
                for item in tldr.split("\n"):
                    print(f"{indent * 2} {item}")

        if explanation_depth > "1":
            if detail:
                print()
                for item in detail.split("\n"):
                    print(f"{indent * 2} {item}")

        print("\n")

    allowed = {str(i) for i in range(len(values))} | {""}
    if allow_stored is not None:
        allowed |= {"s"}
        print(f"{indent[:-1]}[s] Use the stored value {allow_stored}")
        print()

    choice = get_input("", allowed)
    if choice == "":
        choice = str(default)

    elif choice == "s" and allow_stored is not None:
        return str(allow_stored)

    return choice


def authentication_prompt() -> None:
    prev_choice = config_helper.get_user()

    if prev_choice is not None:
        prev_choice = User.sanitize_name(prev_choice)

    if prev_choice == "":
        prev_choice = None

    choice = generic_prompt("There are three ways of storing your password.", [
        ("Encrypted in the database", "Uses the password itself or an optional, additional, password to encrypt your login information.",
         "If you like entering your password this option is for you."),
        ("Clear text in the database", "Your login information is stored in a SQLite Database.",
         f"The password is stored in {path(database_file_location)}\n"
         "It is pretty hard to find and then programmatically extract passwords from a database.\n"
         "So your passwords should be safe."),
        ("Manually entering the information every time", "No passwords are stored. Enter your username every time on demand.", "Use this when you want maximum security and are paranoid."),
    ], default=0, overwrite_output="", allow_stored=prev_choice)

    if choice == "2" or choice == prev_choice:
        return

    print("Please provide authentication for ISIS.")
    username = input("Username: ")
    password = getpass("Password: ")

    config_helper.set_user(username)
    if choice == "0":
        print()
        enc_password = getpass("Additional Password (may be empty): ")
        if enc_password == "":
            enc_password = password

        password = encryptor(enc_password, password)
        config_helper.set_encrypted_password(password)

    else:
        config_helper.set_clear_password(password)


def filename_prompt() -> None:
    prev_choice = config_helper.get_filename_scheme()

    forbidden_chars = "/"
    if is_windows:
        forbidden_chars = "<>:\"/\\|?*"

    choice = generic_prompt(r"""Some programs and programming languages have restrictions
or inconveniences when working with specific characters.

To combat this you can enable a "safe"-mode for the file names.


Once enabled it is not possible to switch back without re-downloading every file.
""", [
        ("No replacing.", f"All characters except {forbidden_chars!r} are left as they are.", ""),
        ("Delete all special characters.",
         "Only ASCII letters + digits + \".\" + \"-\" are allowed. .\n", "When deleting spaces the next character is capitalized."),

    ], default=int(config_helper.default_filename_scheme()), overwrite_output="", allow_stored=prev_choice)

    config_helper.set_filename_scheme(choice)

    print()


def throttler_prompt() -> None:
    prev_choice = config_helper.get_throttle_rate()

    choice = generic_prompt("""If you wish you can throttle your download speed to a limit.
Do you want to do so?

Note: You may overwrite this option by setting the `-d, --download-rate` flag.""", [
        ("No", "", ""),
        ("Yes", "", ""),

    ], default=0, overwrite_output="", allow_stored=prev_choice)

    amount = None
    if choice == "1":
        while True:
            print()
            try:
                amount = input("How many MiB/s am I allowed to consume? ")
                break
            except ValueError:
                clear()
                print("I did not quite catch that\n")

    config_helper.set_throttle_rate(amount)

    print()


def cron_prompt() -> None:
    if is_windows:
        return

    from crontab import CronTab

    cron_works = True
    try:
        with CronTab(user=True) as cron:
            values = [
                ("No", "", ""),
                ("1 Hour", "", ""),
                ("24 Hours", "", ""),
            ]
            command = next(cron.find_command("isisdl"), None)
            if command is not None:
                values.append(("No, but remove the Cron-Job", "", ""))

    except Exception:
        cron_works = False
        values = [
            ("No", "", ""),
        ]

    prompt = "[Linux only]\n\nDo you want me to schedule a Cron-Job to run `isisdl` every x hours?"

    if not cron_works:
        prompt += "\n\nERROR: I could not detect a working Cron-installation.\nThis feature is currently limited to Cron only.\nIn the future there might be support for Systemd timers."

    elif config_helper.get_throttle_rate() is None:
        prompt += "\n\nOn the next page there is a option to throttle your download speed.\nIt is recommended that, if you select a Cron-Job also select a throttle rate."

    choice = generic_prompt(prompt, values, default=0, overwrite_output="")

    if choice == "0":
        return

    with CronTab(user=True) as cron:
        command = next(cron.find_command("isisdl"), None)
        if is_testing:
            return

        if choice == "3" and command is not None:
            cron.remove(command)
            print("Cron-Job erased!")
            return

        if command is not None:
            cron.remove(command)

        import isisdl.__main__
        # Use the executable to have the environment "baked into the interpreter"
        job = cron.new(sys.executable + " " + isisdl.__main__.__file__, "isisdl autogenerated", user=True)

        if choice == "1":
            job.setall("@hourly")
        else:
            job.setall("@daily")

    print()


def telemetry_data_prompt() -> None:
    choice = generic_prompt("""In order to ensure that all courses are downloaded correctly I would like to
collect some metadata from your courses, setup and configuration of isisdl.

I've previously relied on assertions and users reporting these assertions on github.
This system is really inconvenient for both parties and wasts a lot of time.

If you allow it, the program `isisdl` will automatically contact a server when it can report something.

I collect the following:
- Wrong blacklisting of urls
- If files are missing upon rediscovery
- If two files have the same size (or for videos the same length)
- Your Platform
- Your configuration
""", [
        ("No", "", ""),
        ("Yes", "", ""),

    ], default=config_helper.default_telemetry(), overwrite_output="")

    config_helper.set_telemetry(choice)


def update_policy_prompt() -> None:
    prev_choice = config_helper.get_update_policy()
    choice = generic_prompt("""Do you want me to auto-install updates when available?""", [
        ("No", "Do not install any updates.", ""),
        ("Install from github", "",
         "The version of Github is by design always more recent than the one on pip.\n"
         "It should have no stability issues since the update is only downloaded once it passes the tests."),
        ("Install from pip", "", "This build should be always working. Usually pushed ~7-14 days after github release."),
    ], default=int(config_helper.default_update_policy()), overwrite_output="", allow_stored=prev_choice)

    config_helper.set_update_policy(choice)


def main() -> None:
    print(f"""I will guide you through a short configuration phase of about 4min.
It is recommended that you read the options carefully.

If you want to {'accept' if is_first_time else 'reset to'} the default press [d] and [enter].
Otherwise just press [enter].
    """)
    choice = input("")

    if choice.lower() == "d":
        print(f"\n{'Accepted' if is_first_time else 'Reset to'} the defaults!")
        return

    authentication_prompt()
    filename_prompt()
    cron_prompt()
    throttler_prompt()
    update_policy_prompt()
    telemetry_data_prompt()

    print("Thank you for your time - everything is saved!")


if __name__ == "__main__":
    main()
